# Copyright Notice:
# Copyright 2016-2024 DMTF. All rights reserved.
# License: BSD 3-Clause License. For full text see link: https://github.com/DMTF/Redfish-Service-Validator/blob/main/LICENSE.md
#
# Unit tests for RedfishServiceValidator.py
#

import unittest
import sys
import pprint

sys.path.append('../')

import redfish_service_validator.catalog as catalog

import logging

logging.Logger.verbose1 =  logging.Logger.debug
logging.Logger.verbose2 =  logging.Logger.debug

class TestCatalog(unittest.TestCase):
    def test_fuzzy(self):
        print('\n')
        val = catalog.get_fuzzy_property('PropertyA', {'Name': 'Payload', 'PropertyB': False})
        self.assertEqual(val, 'PropertyB')
        val = catalog.get_fuzzy_property('PropertyA', {'Name': 'Payload', 'PropertyB': False, 'PropertyA': False})
        self.assertEqual(val, 'PropertyA')
        val = catalog.get_fuzzy_property('PropertyA', {'Name': 'Payload'}, [])
        self.assertEqual(val, 'PropertyA')
        val = catalog.get_fuzzy_property('PropertyA', {'Name': 'Payload'}, ['PropertyB'])
        self.assertEqual(val, 'PropertyA')
        # OK

    def test_catalog(self):
        print('\n')
        my_catalog = catalog.SchemaCatalog('./tests/testdata/schemas/')

        my_schema_doc = my_catalog.getSchemaDocByClass('Example')

        self.assertRaises(catalog.MissingSchemaError, my_catalog.getSchemaDocByClass, 'NotExample')

        my_schema = my_catalog.getSchemaInCatalog('Example.v1_0_0')

        my_type = my_catalog.getTypeInCatalog('Example.v1_7_0.Example')

        my_type = my_catalog.getTypeInCatalog('Example.v1_2_0.Links')
        # OK
    
    def test_schema_doc(self):
        print('\n')
        my_catalog = catalog.SchemaCatalog('./tests/testdata/schemas/')
        with open('./tests/testdata/schemas/Example_v1.xml') as f:
            my_doc = catalog.SchemaDoc(f.read(), my_catalog, 'Example_v1.xml')
        
        ref = my_doc.getReference('ExampleResource')
        ref = my_doc.getReference('ExampleResource.v1_0_0')
        ref = my_doc.getReference('ExampleResource.v1_0_1')
        ref = my_doc.getReference('ExampleResource.v1_9_9')
        ref = my_doc.getReference('Redfish')
        ref = my_doc.getReference('RedfishExtension.v1_0_0')
    
        my_type = my_doc.getTypeInSchemaDoc('Example.v1_0_0.Example')
        my_type = my_doc.getTypeInSchemaDoc(my_type)
        my_type = my_doc.getTypeInSchemaDoc('Example.v1_9_9.Example')
        my_type = my_doc.getTypeInSchemaDoc('Example.v1_0_0.Actions')
        my_type = my_doc.getTypeInSchemaDoc('ExampleResource.v1_0_0.ExampleResource')
        self.assertRaises(catalog.MissingSchemaError, my_doc.getTypeInSchemaDoc, 'NoExample.v1_0_0.NoExample')

    def test_schema_class(self):
        print('\n')
        my_catalog = catalog.SchemaCatalog('./tests/testdata/schemas/')
        my_doc = my_catalog.getSchemaDocByClass('Example.v1_0_0')
        my_schema = my_catalog.getSchemaInCatalog('Example.v1_0_0')
    
    def test_basic_properties(self):
        print('\nTesting basic types as json')
        prop = catalog.RedfishProperty("Edm.Int").populate(1)
        print(prop.as_json())
        prop = catalog.RedfishProperty("Edm.Decimal").populate(1.1)
        print(prop.as_json())
        prop = catalog.RedfishProperty("Edm.Guid").populate("123")
        print(prop.as_json())
        prop = catalog.RedfishProperty("Edm.Guid").populate(catalog.REDFISH_ABSENT)
        print(prop.as_json())

    def test_basic_properties_check(self):
        print('\nTesting check values')
        prop = catalog.RedfishProperty("Edm.Int").populate(1, check=True)
        prop = catalog.RedfishProperty("Edm.Int").populate(1.1, check=True)
        prop = catalog.RedfishProperty("Edm.Int").populate("1", check=True)
        prop = catalog.RedfishProperty("Edm.Decimal").populate(1.1, check=True)
        prop = catalog.RedfishProperty("Edm.Decimal").populate("1.1", check=True)
        prop = catalog.RedfishProperty("Edm.String").populate("1", check=True)
        prop = catalog.RedfishProperty("Edm.String").populate(1, check=True)
        prop = catalog.RedfishProperty("Edm.Guid").populate("123", check=True)
        prop = catalog.RedfishProperty("Edm.Guid").populate(catalog.REDFISH_ABSENT, check=True)
    
    def test_object(self):
        print('\nTesting object values')
        my_catalog = catalog.SchemaCatalog('./tests/testdata/schemas/')
        my_schema_doc = my_catalog.getSchemaDocByClass("ExampleResource.v1_0_0.ExampleResource")
        my_type = my_schema_doc.getTypeInSchemaDoc("ExampleResource.v1_0_0.ExampleResource")
        object = catalog.RedfishObject( my_type )
        pprint.pprint(object.as_json(), indent=2)
        dct = object.as_json()
        dct = object.getLinks()
        object = catalog.RedfishObject( my_type ).populate({"Id": None, "Description": None})
        pprint.pprint(object.as_json(), indent=2)
        dct = object.as_json()
        dct = object.getLinks()

    def test_capabilities(self):
        my_catalog = catalog.SchemaCatalog('./tests/testdata/schemas/')
        my_schema_doc = my_catalog.getSchemaDocByClass("Example.v1_0_0.Example")
        my_type = my_schema_doc.getTypeInSchemaDoc("Example.v1_0_0.Example")
        my_capabilities = my_type.getCapabilities()

        self.assertTrue(my_capabilities['CanUpdate'])
        self.assertFalse(my_capabilities['CanInsert'])
        self.assertFalse(my_capabilities['CanDelete'])

        self.assertTrue(my_type.CanUpdate)
        self.assertFalse(my_type.CanInsert)
        self.assertFalse(my_type.CanDelete)
    
    def test_expected_uris(self):
        print('\nTesting expected Uris')
        my_catalog = catalog.SchemaCatalog('./tests/testdata/schemas/')
        my_schema_doc = my_catalog.getSchemaDocByClass("Example.v1_0_0.Example")
        my_type = my_schema_doc.getTypeInSchemaDoc("Example.v1_0_0.Example")
        object = catalog.RedfishObject( my_type )

        arr = object.Type.getUris()
        self.assertEqual(len(arr), 3)


        object = catalog.RedfishObject( my_type ).populate({
            "@odata.id": "/redfish/v1/Example",
            "Id": "Example",
            "Description": None
            })
        self.assertTrue(object.HasValidUri)

        object = catalog.RedfishObject( my_type ).populate({
            "@odata.id": "/redfish/v1/Examples",
            "Id": "Examples",
            "Description": None
            })
        self.assertFalse(object.HasValidUri)

        object = catalog.RedfishObject( my_type ).populate({
            "@odata.id": "/redfish/v1/Examples/FunnyId",
            "Id": 'FunnyId',
            "Description": None
            })
        self.assertTrue(object.HasValidUri)
        self.assertTrue(object.HasValidUriStrict)

        object = catalog.RedfishObject( my_type ).populate({
            "@odata.id": "/redfish/v1/Examples/SubObject/FunnyId",
            "Id": 'FunnyId',
            "Description": None
            })
        self.assertTrue(object.HasValidUri)
        self.assertTrue(object.HasValidUriStrict)

        object = catalog.RedfishObject( my_type ).populate({
            "@odata.id": "/redfish/v1/Examples/SubObject/FunnyId",
            "Description": None
            })
        self.assertTrue(object.HasValidUri)
        self.assertTrue(object.HasValidUriStrict)

        object = catalog.RedfishObject( my_type ).populate({
            "@odata.id": "/redfish/v1/Examples/WrongId",
            "Id": 'FunnyId',
            "Description": None
            })
        self.assertTrue(object.HasValidUri)
        self.assertFalse(object.HasValidUriStrict)

        object = catalog.RedfishObject( my_type ).populate({
            "@odata.id": "/redfish/v1/Examples/NoId",
            "Id": None,
            "Description": None
            })
        self.assertTrue(object.HasValidUri)
        self.assertTrue(object.HasValidUriStrict)



if __name__ == '__main__':
    unittest.main()


# Copyright Notice:
# Copyright 2016-2024 DMTF. All rights reserved.
# License: BSD 3-Clause License. For full text see link: https://github.com/DMTF/Redfish-Service-Validator/blob/main/LICENSE.md

import json
from datetime import datetime
from functools import lru_cache
from urllib.parse import urlparse, urlunparse
from http.client import responses
import os
import re

import redfish as rf
import requests
import redfish_service_validator.catalog as catalog
from redfish_service_validator.helper import navigateJsonFragment, splitVersionString
from redfish_service_validator.metadata import Metadata

import logging
my_logger = logging.getLogger(__name__)
traverseLogger = my_logger

CACHE_SIZE = 128

# dictionary to hold sampling notation strings for URIs
class AuthenticationError(Exception):
    """Exception used for failed basic auth or token auth"""
    def __init__(self, msg=None):
        super(AuthenticationError, self).__init__(msg)

def getLogger():
    """
    Grab logger for tools that might use this lib
    """
    return my_logger

class rfService():
    def __init__(self, config):
        traverseLogger.info('Setting up service...')
        self.active, self.config = False, config
        self.logger = getLogger()
    
        self.cache, self.cache_order = {}, []

        self.config['configuri'] = self.config['ip']
        self.config['metadatafilepath'] = self.config['schema_directory']
        self.config['usessl'] = urlparse(self.config['configuri']).scheme in ['https']
        self.config['certificatecheck'] = False
        self.config['certificatebundle'] = None
        self.config['timeout'] = 10

        # NOTE: this is a validator limitation.  maybe move this to its own config
        if self.config['collectionlimit']:
            total_len = len(self.config['collectionlimit']) / 2
            limit_string = ' '.join(self.config['collectionlimit'])
            limit_array = [tuple(found_item.split(' ')) for found_item in re.findall(r"[A-Za-z]+ [0-9]+", limit_string)]
            if len(limit_array) != total_len:
                raise ValueError('Collection Limit array seems malformed, use format: RESOURCE1 COUNT1 RESOURCE2 COUNT2)...')
            self.config['collectionlimit'] = {x[0]: int(x[1]) for x in limit_array}

        # Log into the service
        if not self.config['usessl'] and not self.config['forceauth']:
            if self.config['username'] not in ['', None] or self.config['password'] not in ['', None]:
                traverseLogger.warning('Attempting to authenticate on unchecked http/https protocol is insecure, if necessary please use ForceAuth option.  Clearing auth credentials...')
                self.config['username'] = ''
                self.config['password'] = ''
        rhost, user, passwd = self.config['configuri'], self.config['username'], self.config['password']
        proxies=None
        if self.config['serv_http_proxy'] != '' or self.config['serv_https_proxy'] != '':
            proxies = {}
            if self.config['serv_http_proxy'] != '': proxies['http'] = self.config['serv_http_proxy']
            if self.config['serv_https_proxy'] != '': proxies['https'] = self.config['serv_https_proxy']
        self.ext_proxies=None
        if self.config['ext_http_proxy'] != '' or self.config['ext_https_proxy'] != '':
            self.ext_proxies = {}
            if self.config['ext_http_proxy'] != '': self.ext_proxies['http'] = self.config['ext_http_proxy']
            if self.config['ext_https_proxy'] != '': self.ext_proxies['https'] = self.config['ext_https_proxy']
        self.context = rf.redfish_client(base_url=rhost, username=user, password=passwd, timeout=self.config['timeout'], proxies=proxies)
        self.context.login( auth = self.config['authtype'].lower() )

        # Go through $metadata and download any additional schema files needed
        success, data, response, delay = self.callResourceURI(Metadata.metadata_uri)
        if success and data is not None and response.status in range(200,210):
            self.metadata = Metadata(data, self, my_logger)
        else:
            self.metadata = Metadata(None, self, my_logger)

        # Build the data model based on cached schema files
        self.catalog = catalog.SchemaCatalog(self.config['metadatafilepath'])

        target_version = 'n/a'

        # get Version
        success, data, response, delay = self.callResourceURI('/redfish/v1')
        if not success:
            traverseLogger.warning('Could not get ServiceRoot')
        else:
            if 'RedfishVersion' not in data:
                traverseLogger.warning('Could not get RedfishVersion from ServiceRoot')
            else:
                traverseLogger.info('Redfish Version of Service: {}'.format(data['RedfishVersion']))
                target_version = data['RedfishVersion']
        if target_version in ['1.0.0', 'n/a']:
            traverseLogger.warning('!!Version of target may produce issues!!')
        if splitVersionString(target_version) < splitVersionString('1.6.0') and not self.config['uricheck']:
            traverseLogger.warning('RedfishVersion below 1.6.0, disabling uri checks')
            self.catalog.flags['ignore_uri_checks'] = True
        else:
            self.catalog.flags['ignore_uri_checks'] = False
            self.config['uricheck'] = True

        self.service_root = data

        self.active = True

    def close(self):
        self.active = False
    
    def callResourceURI(self, link_uri):
        traverseLogger = my_logger
        """
        Makes a call to a given URI or URL

        param arg1: path to URI "/example/1", or URL "http://example.com"
        return: (success boolean, data, request status code)
        """
        # rs-assertions: 6.4.1, including accept, content-type and odata-versions
        # rs-assertion: handle redirects?  and target permissions
        # rs-assertion: require no auth for serviceroot calls
        # TODO: Written with "success" values, should replace with Exception and catches
        if link_uri is None:
            traverseLogger.warning("This URI is empty!")
            return False, None, None, 0
        
        config = self.config
        # proxies = self.proxies
        ConfigIP, UseSSL, AuthType, ChkCert, ChkCertBundle, timeout, Token = config['configuri'], config['usessl'], config['authtype'], \
                config['certificatecheck'], config['certificatebundle'], config['timeout'], config['token']

        scheme, netloc, path, params, query, fragment = urlparse(link_uri)
        inService = scheme == '' and netloc == ''
        if inService:
            my_destination = urlunparse((scheme, netloc, path, '', '', ''))  # URILink
        else:
            my_destination = urlunparse((scheme, netloc, path, params, query, fragment))

        payload, statusCode, elapsed, auth, noauthchk = None, '', 0, None, True

        isXML = False
        if "$metadata" in path or ".xml" in path[:-5]:
            isXML = True
            traverseLogger.debug('Should be XML')

        # determine if we need to Auth...
        if inService:
            noauthchk = link_uri in ['/redfish', '/redfish/v1', '/redfish/v1/odata'] or\
                '/redfish/v1/$metadata' in link_uri

            auth = None if noauthchk else (config.get('username'), config.get('password'))
            traverseLogger.debug('dont chkauth' if noauthchk else 'chkauth')

        # rs-assertion: do not send auth over http
        # remove UseSSL if necessary if you require unsecure auth
        if (not UseSSL and not config['forceauth']) or not inService or AuthType != 'Basic':
            auth = None

        # only send token when we're required to chkauth, during a Session, and on Service and Secure
        headers = {"Accept-Encoding": "*"}

        certVal = ChkCertBundle if ChkCert and ChkCertBundle not in [None, ""] else ChkCert

        # rs-assertion: must have application/json or application/xml
        traverseLogger.debug('callingResourceURI {}with authtype {} and ssl {}: {} {}'.format(
            'out of service ' if not inService else '', AuthType, UseSSL, link_uri, headers))
        response = None
        try:
            startTick = datetime.now()
            if my_destination not in self.cache:
                mockup_file_path = os.path.join(config['mockup'], my_destination.replace('/redfish/v1/', '', 1).strip('/'), 'index.json')
                if not inService:
                    req = requests.get(my_destination, proxies=self.ext_proxies, verify=False)
                    content = req.json if not isXML else req.text
                    response = rf.rest.v1.StaticRestResponse(Status=req.status_code, Headers={x:req.headers[x] for x in req.headers}, Content=req.text)
                elif config['mockup'] != '' and os.path.isfile(mockup_file_path):
                    content = {}
                    with open(mockup_file_path) as mockup_file:
                        content = json.load(mockup_file)
                    response = rf.rest.v1.StaticRestResponse(Status=200, Headers={'Content-Type': 'application/json', 'X-Redfish-Mockup': 'true'}, Content=content)
                else:
                    response = self.context.get(my_destination, headers=headers)
                self.cache[my_destination] = response
                self.cache_order.append(my_destination)
                if len(self.cache) > CACHE_SIZE:
                    del self.cache[self.cache_order.pop(0)]
            else:
                traverseLogger.debug("CACHE HIT {} {}".format(my_destination, link_uri))
            
            response = self.cache[my_destination]

            elapsed = datetime.now() - startTick
            statusCode = response.status

            traverseLogger.debug('{}, {},\nTIME ELAPSED: {}'.format(statusCode, response.getheaders(), elapsed))
            if statusCode in [200]:
                contenttype = response.getheader('content-type')
                if contenttype is None:
                    traverseLogger.error("Content-type not found in header: {}".format(link_uri))
                    contenttype = ''
                if 'application/json' in contenttype:
                    traverseLogger.debug("This is a JSON response")
                    decoded = response.dict
                            
                    # navigate fragment
                    decoded = navigateJsonFragment(decoded, link_uri)
                    if decoded is None:
                        traverseLogger.error(
                                "The JSON pointer in the fragment of this URI is not constructed properly: {}".format(link_uri))
                elif 'application/xml' in contenttype:
                    decoded = response.text
                elif 'text/xml' in contenttype:
                    # non-service schemas can use "text/xml" Content-Type
                    if inService:
                        traverseLogger.warning(
                                "Incorrect content type 'text/xml' for file within service {}".format(link_uri))
                    decoded = response.text
                else:
                    traverseLogger.error(
                            "This URI did NOT return XML or Json contenttype, is this not a Redfish resource (is this redirected?): {}".format(link_uri))
                    decoded = None
                    if isXML:
                        traverseLogger.info('Attempting to interpret as XML')
                        decoded = response.text
                    else:
                        try:
                            json.loads(response.text)
                            traverseLogger.info('Attempting to interpret as JSON')
                            decoded = response.dict
                        except ValueError:
                            pass

                return decoded is not None, decoded, response, elapsed
            elif statusCode == 401:
                if inService and AuthType in ['Basic', 'Token']:
                    if AuthType == 'Token':
                        cred_type = 'token'
                    else:
                        cred_type = 'username and password'
                    raise AuthenticationError('Error accessing URI {}. Status code "{} {}". Check {} supplied for "{}" authentication.'
                                              .format(link_uri, statusCode, responses[statusCode], cred_type, AuthType))

        except AuthenticationError as e:
            raise e  # re-raise exception
        except Exception as e:
            traverseLogger.error("A problem when getting resource {} has occurred: {}".format(link_uri, repr(e)))
            traverseLogger.debug("output: ", exc_info=True)
            if response and response.text:
                traverseLogger.debug("payload: {}".format(response.text))

        if payload is not None:
            return True, payload, response, 0
        return False, None, response, elapsed



# Copyright Notice:
# Copyright 2016-2024 DMTF. All rights reserved.
# License: BSD 3-Clause License. For full text see link: https://github.com/DMTF/Redfish-Service-Validator/blob/main/LICENSE.md

import argparse
import os
import logging
from io import BytesIO
import zipfile
import requests

# live_zip_uri = 'http://redfish.dmtf.org/schemas/DSP8010_2021.1.zip'
live_zip_uri = 'https://www.dmtf.org/sites/default/files/standards/documents/DSP8010.zip' 

my_logger = logging.getLogger()

def setup_schema_pack(uri, local_dir, http_proxy='', https_proxy=''):
    proxies, timeout = None, 20
    if http_proxy != '' or https_proxy != '':
        proxies = {}
        if http_proxy != '': proxies['http'] = http_proxy
        if https_proxy != '': proxies['https'] = https_proxy
    if uri == 'latest':
        uri = live_zip_uri
    my_logger.info('Unpacking schema pack... {}'.format(uri))
    try:
        if not os.path.isdir(local_dir):
            os.makedirs(local_dir)
        response = requests.get(uri, timeout=timeout, proxies=proxies)
        expCode = [200]
        elapsed = response.elapsed.total_seconds()
        statusCode = response.status_code
        my_logger.debug('{}, {}, {},\nTIME ELAPSED: {}'.format(statusCode,
                                                                        expCode, response.headers, elapsed))
        if statusCode in expCode:
            if not zipfile.is_zipfile(BytesIO(response.content)):
                my_logger.error('This URL did not return a valid zipfile')
                pass
            else:
                zf = zipfile.ZipFile(BytesIO(response.content))
                zf.testzip()
                for name in zf.namelist():
                    if '.xml' in name:
                        cpath = '{}/{}'.format(local_dir, name.split('/')[-1])
                        my_logger.debug((name, cpath))
                        item = zf.open(name)
                        with open(cpath, 'wb') as f:
                            f.write(item.read())
                        item.close()
                zf.close()
    except Exception as ex:
        my_logger.error("A problem when getting resource has occurred {}".format(uri))
        my_logger.warning("output: ", exc_info=True)
    return True


if __name__ == '__main__':
    argget = argparse.ArgumentParser(description='Acquire schema_pack from DMTF website')

    # config
    argget.add_argument('--source', type=str, default=live_zip_uri, help='URL of the given schemapack, if unspecified, always grab latest')
    argget.add_argument('--schema_directory', type=str, default='./SchemaFiles/metadata', help='directory for local schema files')

    args = argget.parse_args()

    setup_schema_pack(args.source, args.schema_directory)



# Copyright Notice:
# Copyright 2016-2024 DMTF. All rights reserved.
# License: BSD 3-Clause License. For full text see link: https://github.com/DMTF/Redfish-Service-Validator/blob/main/LICENSE.md

import os
import sys
import argparse
import logging
import json
from datetime import datetime
import traceback
from redfish_service_validator.metadata import getSchemaDetails
from redfish_service_validator.config import convert_config_to_args, convert_args_to_config
from redfish_service_validator.validateResource import validateSingleURI, validateURITree
from redfish_service_validator import tohtml, schema_pack, traverse
from urllib.parse import urlparse
from collections import Counter

tool_version = '2.4.8'

# Set up the custom debug levels
VERBOSE1 = logging.INFO-1
VERBOSE2 = logging.INFO-2

logging.addLevelName(VERBOSE1, "VERBOSE1")
logging.addLevelName(VERBOSE2, "VERBOSE2")


def print_verbose_1(self, msg, *args, **kwargs):
    if self.isEnabledFor(VERBOSE1):
        self._log(VERBOSE1, msg, args, **kwargs)


def print_verbose_2(self, msg, *args, **kwargs):
    if self.isEnabledFor(VERBOSE2):
        self._log(VERBOSE2, msg, args, **kwargs)
        

logging.Logger.verbose1 = print_verbose_1
logging.Logger.verbose2 = print_verbose_2

my_logger = logging.getLogger()
my_logger.setLevel(logging.DEBUG)
standard_out = logging.StreamHandler(sys.stdout)
standard_out.setLevel(logging.INFO)
my_logger.addHandler(standard_out)


def validate(argslist=None, configfile=None):
    """Main command

    Args:
        argslist ([type], optional): List of arguments in the form of argv. Defaults to None.
    """    
    argget = argparse.ArgumentParser(description='DMTF tool to test a service against a collection of Schema, version {}'.format(tool_version))

    # base tool
    argget.add_argument('-v', '--verbose', action='count', default=0, help='Verbosity of tool in stdout')
    argget.add_argument('-c', '--config', type=str, help='Configuration for this tool')

    # host info
    argget.add_argument('-i', '--ip', '--rhost', '-r', type=str, help='The address of the Redfish service (with scheme); example: \'https://123.45.6.7:8000\'')
    argget.add_argument('-u', '--username', '-user', type=str, help='The username for authentication')
    argget.add_argument('-p', '--password', type=str, help='The password for authentication')
    argget.add_argument('--description', type=str, help='The description of the system for identifying logs; if none is given, a value is produced from information in the service root')
    argget.add_argument('--forceauth', action='store_true', help='Force authentication on unsecure connections')
    argget.add_argument('--authtype', type=str, default='Basic', help='Authorization type; \'None\', \'Basic\', \'Session\', or \'Token\'')
    argget.add_argument('--token', type=str, help='Token when \'authtype\' is \'Token\'')
    argget.add_argument('--ext_http_proxy', type=str, default='', help='URL of the HTTP proxy for accessing external sites')
    argget.add_argument('--ext_https_proxy', type=str, default='', help='URL of the HTTPS proxy for accessing external sites')
    argget.add_argument('--serv_http_proxy', type=str, default='', help='URL of the HTTP proxy for accessing the service')
    argget.add_argument('--serv_https_proxy', type=str, default='', help='URL of the HTTPS proxy for accessing the service')

    # validator options
    argget.add_argument('--payload', type=str, help='The mode to validate payloads (\'Tree\', \'Single\', \'SingleFile\', or \'TreeFile\') followed by resource/filepath', nargs=2)
    argget.add_argument('--logdir', '--report-dir', type=str, default='./logs', help='The directory for generated report files; default: \'logs\'')
    argget.add_argument('--nooemcheck', action='store_false', dest='oemcheck', help='Don\'t check OEM items')
    argget.add_argument('--debugging', action="store_true", help='Output debug statements to text log, otherwise it only uses INFO')
    argget.add_argument('--uricheck', action="store_true", help='Allow URI checking on services below RedfishVersion 1.6.0')
    argget.add_argument('--schema_directory', type=str, default='./SchemaFiles/metadata', help='Directory for local schema files')
    argget.add_argument('--mockup', type=str, default='', help='Enables insertion of local mockup resources to replace missing, incomplete, or incorrect implementations retrieved from the service that may hinder full validation coverage')
    argget.add_argument('--collectionlimit', type=str, default=['LogEntry', '20'], help='apply a limit to collections (format: RESOURCE1 COUNT1 RESOURCE2 COUNT2...)', nargs='+')

    # parse...
    args = argget.parse_args(argslist)

    if configfile is None:
        configfile = args.config

    # set logging file
    startTick = datetime.now()

    standard_out.setLevel(logging.INFO - args.verbose if args.verbose < 3 else logging.DEBUG)

    logpath = args.logdir

    if not os.path.isdir(logpath):
        os.makedirs(logpath)

    fmt = logging.Formatter('%(levelname)s - %(message)s')
    file_handler = logging.FileHandler(datetime.strftime(startTick, os.path.join(logpath, "ConformanceLog_%m_%d_%Y_%H%M%S.txt")))
    file_handler.setLevel(min(logging.INFO if not args.debugging else logging.DEBUG, standard_out.level))
    file_handler.setFormatter(fmt)
    my_logger.addHandler(file_handler)

    # begin logging
    my_logger.info("Redfish Service Validator, version {}".format(tool_version))
    my_logger.info("")

    # config verification
    if args.ip is None and configfile is None:
        my_logger.error('No IP or Config Specified')
        argget.print_help()
        return 1, None, 'Configuration Incomplete'

    if configfile:
        convert_config_to_args(args, configfile)
    else:
        my_logger.info('Writing config file to log directory')
        configfilename = datetime.strftime(startTick, os.path.join(logpath, "ConfigFile_%m_%d_%Y_%H%M%S.ini"))
        my_config = convert_args_to_config(args)
        with open(configfilename, 'w') as f:
            my_config.write(f)

    scheme, netloc, path, params, query, fragment = urlparse(args.ip)
    if scheme not in ['http', 'https', 'http+unix']:
        my_logger.error('IP is missing http or https or http+unix')
        return 1, None, 'IP Incomplete'

    if netloc == '':
        my_logger.error('IP is missing ip/host')
        return 1, None, 'IP Incomplete'

    if len(args.collectionlimit) % 2 != 0:
        my_logger.error('Collection Limit requires two arguments per entry (ResourceType Count)')
        return 1, None, 'Collection Limit Incomplete'

    # start printing config details, remove redundant/private info from print
    my_logger.info('Target URI: ' + args.ip)
    my_logger.info('\n'.join(
        ['{}: {}'.format(x, vars(args)[x] if x not in ['password'] else '******') for x in sorted(list(vars(args).keys() - set(['description']))) if vars(args)[x] not in ['', None]]))
    my_logger.info('Start time: ' + startTick.strftime('%x - %X'))
    my_logger.info("")

    # schema and service init
    schemadir = args.schema_directory

    if not os.path.isdir(schemadir):
        my_logger.info('Downloading initial schemas from online')
        my_logger.info('The tool will, by default, attempt to download and store XML files to relieve traffic from DMTF/service')
        schema_pack.my_logger.addHandler(file_handler)
        schema_pack.setup_schema_pack('latest', args.schema_directory, args.ext_http_proxy, args.ext_https_proxy)

    try:
        currentService = traverse.rfService(vars(args))
    except Exception as ex:
        traceback.print_exc()
        my_logger.verbose1('Exception caught while creating Service', exc_info=1)
        my_logger.error("Service could not be started: {}".format(repr(ex)))
        my_logger.error("Try running the Redfish Protocol Validator to ensure the service meets basic protocol conformance")
        return 1, None, 'Service Exception'
    
    if args.description is None and currentService.service_root:
        my_version = currentService.service_root.get('RedfishVersion', 'No Version')
        my_name = currentService.service_root.get('Name', '')
        my_uuid = currentService.service_root.get('UUID', 'No UUID')
        setattr(args, 'description', 'My Target System {}, version {}, {}'.format(my_name, my_version, my_uuid))
    
    my_logger.info('Description of service: {}'.format(args.description))

    # Start main
    status_code = 1
    jsonData = None

    if args.payload:
        pmode, ppath = args.payload
    else:
        pmode, ppath = 'Default', ''
    pmode = pmode.lower()

    if pmode not in ['tree', 'single', 'singlefile', 'treefile', 'default']:
        pmode = 'Default'
        my_logger.error('PayloadMode or path invalid, using Default behavior')
    if 'file' in pmode:
        if ppath is not None and os.path.isfile(ppath):
            with open(ppath) as f:
                jsonData = json.load(f)
                f.close()
        else:
            my_logger.error('File not found for payload: {}'.format(ppath))
            return 1, None, 'File not found for payload: {}'.format(ppath)
    try:
        if 'single' in pmode:
            success, counts, results, xlinks, topobj = validateSingleURI(currentService, ppath, 'Target', expectedJson=jsonData)
        elif 'tree' in pmode:
            success, counts, results, xlinks, topobj = validateURITree(currentService, ppath, 'Target', expectedJson=jsonData)
        else:
            success, counts, results, xlinks, topobj = validateURITree(currentService, '/redfish/v1/', 'ServiceRoot', expectedJson=jsonData)
    except traverse.AuthenticationError as e:
        # log authentication error and terminate program
        my_logger.error('{}'.format(e))
        return 1, None, 'Failed to authenticate with the service'

    currentService.close()

    # get final counts
    metadata = currentService.metadata
    my_logger.verbose1('\nMetadata: Namespaces referenced in service: {}'.format(metadata.get_service_namespaces()))
    my_logger.info('Metadata: Namespaces missing from $metadata: {}'.format(metadata.get_missing_namespaces()))

    if len(metadata.get_missing_namespaces()) > 0:
        my_logger.error('Metadata is missing Namespaces that are referenced by the service.')

    finalCounts = Counter()
    nowTick = datetime.now()
    my_logger.info('\nElapsed time: {}'.format(str(nowTick-startTick).rsplit('.', 1)[0]))

    error_lines, finalCounts = tohtml.count_errors(results)

    for line in error_lines:
        my_logger.error(line)

    finalCounts.update(metadata.get_counter())

    fails = 0
    for key in [key for key in finalCounts.keys()]:
        if finalCounts[key] == 0:
            del finalCounts[key]
            continue
        if any(x in key for x in ['problem', 'fail', 'bad', 'exception']):
            fails += finalCounts[key]

    html_str = tohtml.renderHtml(results, tool_version, startTick, nowTick, currentService)

    lastResultsPage = datetime.strftime(startTick, os.path.join(logpath, "ConformanceHtmlLog_%m_%d_%Y_%H%M%S.html"))

    tohtml.writeHtml(html_str, lastResultsPage)

    success = success and not (fails > 0)
    my_logger.info("\n".join('{}: {}   '.format(x, y) for x, y in sorted(finalCounts.items())))

    # dump cache info to debug log
    my_logger.debug('getSchemaDetails() -> {}'.format(getSchemaDetails.cache_info()))
    my_logger.debug('callResourceURI() -> {}'.format(currentService.cache_order))

    if not success:
        my_logger.error("Validation has failed: {} problems found".format(fails))
    else:
        my_logger.info("Validation has succeeded.")
        status_code = 0

    return status_code, lastResultsPage, 'Validation done'

def main():
    """
    Entry point for the program.
    """
    status_code, _, _ = validate()
    return status_code

if __name__ == '__main__':
    try:
        sys.exit(main())
    except Exception as e:
        my_logger.exception("Program finished prematurely: %s", e)
        raise



# Copyright Notice:
# Copyright 2016-2024 DMTF. All rights reserved.
# License: BSD 3-Clause License. For full text see link: https://github.com/DMTF/Redfish-Service-Validator/blob/main/LICENSE.md

import logging
from collections import Counter, OrderedDict
from io import StringIO

import redfish_service_validator.traverse as traverse
import redfish_service_validator.catalog as catalog
from redfish_service_validator.validateRedfish import checkPropertyConformance, displayValue
from redfish_service_validator.helper import getNamespace, getType, createContext, checkPayloadConformance, navigateJsonFragment, create_entry

my_logger = logging.getLogger()
my_logger.setLevel(logging.DEBUG)
class WarnFilter(logging.Filter):
       def filter(self, rec):
           return rec.levelno == logging.WARN

fmt = logging.Formatter('%(levelname)s - %(message)s')

def create_logging_capture(this_logger):
    errorMessages = StringIO()
    warnMessages = StringIO()

    errh = logging.StreamHandler(errorMessages)
    errh.setLevel(logging.ERROR)
    errh.setFormatter(fmt)

    warnh = logging.StreamHandler(warnMessages)
    warnh.setLevel(logging.WARN)
    warnh.addFilter(WarnFilter())
    warnh.setFormatter(fmt)

    this_logger.addHandler(errh)
    this_logger.addHandler(warnh)

    return errh, warnh


def get_my_capture(this_logger, handler):
    this_logger.removeHandler(handler)
    strings = handler.stream.getvalue()
    handler.stream.close()
    return strings


def validateSingleURI(service, URI, uriName='', expectedType=None, expectedJson=None, parent=None):
    # rs-assertion: 9.4.1
    # Initial startup here
    my_logger.verbose1("\n*** %s, %s", uriName, URI)
    my_logger.info("\n*** %s", URI)
    my_logger.verbose1("\n*** {}, {}".format(expectedType, expectedJson is not None))
    counts = Counter()
    results, messages = OrderedDict(), OrderedDict()

    ehandler, whandler = create_logging_capture(my_logger)

    me = {'uri': URI, 'success': False, 'counts': counts, 'messages': messages,
            'errors': '', 'warns': '', 'rtime': '', 'rcode': 0,
            'fulltype': '', 'context': '...', 'payload': {}}
    results[uriName] = me

    # check for @odata mandatory stuff
    # check for version numbering problems # check id if its the same as URI
    # check @odata.context instead of local.  Realize that @odata is NOT a "property"

    # Attempt to get a list of properties
    if URI is None:
        URI = '/Missing URI Link'
        if parent: URI = str(parent.payload.get('@odata.id')) + URI
        my_logger.warning('Tool appears to be missing vital URI information, replacing URI w/: {}'.format(URI))
    # Generate dictionary of property info
    try:
        if expectedJson is None:
            ret = service.callResourceURI(URI)
            success, me['payload'], response, me['rtime'] = ret
            me['rcode'] = response.status if response else -1
        else:
            success, me['payload'], me['rcode'], me['rtime'] = True, expectedJson, -1, 0
            response = None
        
        if not success:
            my_logger.error('URI did not return resource {}'.format(URI))
            counts['failGet'] += 1
            # Failure to connect to the scheme is an important error that must be included in FAILS
            counts['fails'] += 1
            me['warns'], me['errors'] = get_my_capture(my_logger, whandler), get_my_capture(my_logger, ehandler)
            return False, counts, results, None, None

        # verify basic odata strings
        if results[uriName]['payload'] is not None:
            successPayload, odataMessages = checkPayloadConformance(me['payload'], URI)
            for m in odataMessages:
                msg = create_entry(m, *odataMessages[m])
                messages[msg.name] = msg

        my_type = me['payload'].get('@odata.type', expectedType)
        me['fulltype'] = str(my_type)
        if my_type is None:
            redfish_obj = None
        else:
            # TODO: don't have the distinction between Property Type and a Normal Type
            if isinstance(my_type, catalog.RedfishType):
                my_type = my_type.fulltype
            redfish_schema = service.catalog.getSchemaDocByClass(my_type)
            redfish_type = redfish_schema.getTypeInSchemaDoc(my_type)

            redfish_obj = catalog.RedfishObject(redfish_type, 'Object', parent=parent).populate(me['payload']) if redfish_type else None

        if redfish_obj:
            me['fulltype'] = redfish_obj.Type.fulltype
        else:
            counts['problemResource'] += 1
            me['warns'], me['errors'] = get_my_capture(my_logger, whandler), get_my_capture(my_logger, ehandler)
            return False, counts, results, None, None
    except traverse.AuthenticationError as e:
        raise  # re-raise exception
    except Exception as e:
        my_logger.verbose1('Exception caught while creating ResourceObj', exc_info=1)
        my_logger.error('Unable to gather property info from schema for URI {}; check its schema definition for schema errors: {}'.format(URI, repr(e)))
        counts['exceptionResource'] += 1
        # ExceptionResource is an important error that must be included in FAILS
        counts['fails'] += 1
        me['warns'], me['errors'] = get_my_capture(my_logger, whandler), get_my_capture(my_logger, ehandler)
        return False, counts, results, None, None

    counts['passGet'] += 1

    # verify odata_id properly resolves to its parent if holding fragment
    odata_id = me['payload'].get('@odata.id')
    if odata_id is None:
        # Do not error for namespace.type MessageRegistry.MessageRegistry, etc 
        if any(['{}.{}'.format(x, x) in redfish_obj.Type.getTypeTree() for x in ['MessageRegistry', 'AttributeRegistry', 'PrivilegeRegistry']]):
            my_logger.debug('No @odata.id was found in this resource, but not needed')
        else:
            my_logger.error('No @odata.id was found in this resource')
            messages['@odata.id'] = create_entry('@odata.id', '-', '-', 'DNE', 'FAIL')
            counts['errorMissingOdataId'] += 1

    if odata_id is not None and '#' in odata_id:
        if parent is not None:
            payload_resolve = navigateJsonFragment(parent.payload, URI)
            if parent.payload.get('@odata.id') not in URI:
                my_logger.info('@odata.id of ReferenceableMember was referenced elsewhere...: {}'.format(odata_id))
            elif payload_resolve is None:
                my_logger.error('@odata.id of ReferenceableMember does not contain a valid JSON pointer for this payload: {}'.format(odata_id))
                counts['badOdataIdResolution'] += 1
            elif payload_resolve != me['payload']:
                my_logger.error('@odata.id of ReferenceableMember does not point to the correct object: {}'.format(odata_id))
                counts['badOdataIdResolution'] += 1
            _, end_fragment = tuple(odata_id.split('#', 1))
            my_member_id = me['payload'].get('MemberId')
            if not my_member_id:
                my_logger.error('ReferenceableMember MemberId does not exist...')
                counts['badMemberId'] += 1
            elif my_member_id not in end_fragment.split('/'):
                my_logger.error('ReferenceableMember MemberId does not match id: {} {}'.format(my_member_id, odata_id))
                counts['badMemberId'] += 1
        else:
            my_logger.warning('No parent found with which to test @odata.id of ReferenceableMember')
    
    if service.config['uricheck']:
        my_uris = redfish_obj.Type.getUris()
        if odata_id is not None and redfish_obj.Populated and len(my_uris) > 0:
            if redfish_obj.HasValidUri:
                counts['passRedfishUri'] += 1
                if not redfish_obj.HasValidUriStrict and redfish_obj.payload.get('Id') is not None:
                    counts['failRedfishUriStrict'] += 1
                    messages['@odata.id'].result = 'FAIL'
                    my_logger.error('The Id property does not match the last segment of the URI {}'.format(odata_id))
            else:
                if '/Oem/' in odata_id:
                    counts['warnRedfishUri'] += 1
                    messages['@odata.id'].result = 'WARN'
                    my_logger.warning('URI {} does not match the following required URIs in Schema of {}'.format(odata_id, redfish_obj.Type))
                else:
                    counts['failRedfishUri'] += 1
                    messages['@odata.id'].result = 'FAIL'
                    my_logger.error('URI {} does not match the following required URIs in Schema of {}'.format(odata_id, redfish_obj.Type))

    if response and response.getheader('Allow'):
        allowed_responses = [x.strip().upper() for x in response.getheader('Allow').split(',')]
        if not redfish_obj.Type.CanInsert and 'POST' in allowed_responses:
            my_logger.error('Allow header should NOT contain POST for {}'.format(redfish_obj.Type))
            counts['failAllowHeader'] += 1
        if not redfish_obj.Type.CanDelete and 'DELETE' in allowed_responses:
            my_logger.error('Allow header should NOT contain DELETE for {}'.format(redfish_obj.Type))
            counts['failAllowHeader'] += 1
        if not redfish_obj.Type.CanUpdate and any([x in allowed_responses for x in ['PATCH', 'PUT']]):
            my_logger.warning('Allow header should NOT contain PATCH or PUT for {}'.format(redfish_obj.Type))
            counts['warnAllowHeader'] += 1

    if response and response.getheader('x-Redfish-Mockup'):
        my_logger.warning('Response payload loaded from mockup, not the service under test')
        counts['warnMockupUsed'] += 1

    if not successPayload:
        counts['failPayloadError'] += 1
        my_logger.error(str(URI) + ': payload error, @odata property non-conformant',)

    # if URI was sampled, get the notation text from traverseService.uri_sample_map
    results[uriName]['uri'] = (str(URI))
    results[uriName]['context'] = createContext(me['fulltype'])
    results[uriName]['origin'] = redfish_obj.Type.owner.parent_doc.name
    results[uriName]['success'] = True

    my_logger.info("\t Type (%s), GET SUCCESS (time: %s)", me['fulltype'], me['rtime'])
    
    for prop_name, prop in redfish_obj.properties.items():
        try:
            if not prop.HasSchema and not prop.Exists:
                counts['skipNoSchema'] += 1
                continue
            elif not prop.HasSchema:
                my_logger.error('No Schema for property {}'.format(prop.Name))
                counts['errorNoSchema'] += 1
                continue
            propMessages, propCounts = checkPropertyConformance(service, prop_name, prop)

            propMessages = {x:create_entry(x, *y) if isinstance(y, tuple) else y for x,y in propMessages.items()}

            if 'MessageRegistry.MessageRegistry' not in redfish_obj.Type.getTypeTree():
                if '@Redfish.Copyright' in propMessages:
                    modified_entry = propMessages['@Redfish.Copyright']
                    modified_entry.result = 'FAIL'
                    my_logger.error('@Redfish.Copyright is only allowed for mockups, and should not be allowed in official implementations')

            messages.update(propMessages)
            counts.update(propCounts)
        except traverse.AuthenticationError as e:
            raise  # re-raise exception
        except Exception as ex:
            my_logger.verbose1('Exception caught while validating single URI', exc_info=1)
            my_logger.error('{}: Could not finish check on this property ({})'.format(prop_name, str(ex)))
            messages[prop_name] = create_entry(prop_name, '', '', '...', 'exception')
            counts['exceptionPropCheck'] += 1

    SchemaFullType, jsonData = me['fulltype'], me['payload']
    SchemaNamespace, SchemaType = getNamespace(SchemaFullType), getType(SchemaFullType)

    # List all items checked and unchecked
    # current logic does not check inside complex types
    fmt = '%-30s%30s'
    my_logger.verbose1('%s, %s, %s', uriName, SchemaNamespace, SchemaType)

    for key in jsonData:
        my_logger.verbose1(fmt % (key, messages[key].result if key in messages else 'Exists, no schema check'))

    allowAdditional = redfish_obj.Type.HasAdditional
    for key in [k for k in jsonData if k not in messages and k not in redfish_obj.properties and '@' not in k]:
        # note: extra messages for "unchecked" properties
        item = jsonData.get(key)
        if not allowAdditional:
            my_logger.error('{} not defined in schema {} (check version, spelling and casing)'.format(key, SchemaNamespace))
            counts['failAdditional'] += 1
            messages[key] = create_entry(key, displayValue(item), '-', '-', 'FAIL')
        else:
            my_logger.warning('{} not defined in schema {} (check version, spelling and casing)'.format(key, SchemaNamespace))
            counts['unverifiedAdditional'] += 1
            messages[key] = create_entry(key, displayValue(item), '-', '-', 'Additional')

        fuzz = catalog.get_fuzzy_property(key, redfish_obj.properties)
        if fuzz != key and fuzz in redfish_obj.properties:
            messages[fuzz] = create_entry(fuzz, '-', '-', '-', 'INVALID')
            my_logger.error('Attempting {} (from {})?'.format(fuzz, key))
            my_new_obj = redfish_obj.properties[fuzz].populate(item)
            new_msgs, new_counts = checkPropertyConformance(service, key, my_new_obj)
            new_msgs = {x:create_entry(x, *y) for x,y in new_msgs.items()}
            messages.update(new_msgs)
            counts.update(new_counts)
            counts['invalidNamedProperty'] += 1

    for key in messages:
        if key not in jsonData:
            my_logger.verbose1(fmt % (key, messages[key].result))

    results[uriName]['warns'], results[uriName]['errors'] = get_my_capture(my_logger, whandler), get_my_capture(my_logger, ehandler)

    pass_val = len(results[uriName]['errors']) == 0
    for key in counts:
        if any(x in key for x in ['problem', 'fail', 'bad', 'exception']):
            pass_val = False
            break
    my_logger.info("\t {}".format('PASS' if pass_val else ' FAIL...'))

    my_logger.verbose1('%s, %s', SchemaFullType, counts)

    # Get all links available

    collection_limit = service.config['collectionlimit']

    # Count of occurrences of fail, warn, invalid and deprecated in result of tests to FAILS / WARNINGS
    for value in messages.values():
        if "FAIL" in value.result:
            counts['fails'] += 1
        if "WARN" in value.result or "INVALID" in value.result or "Deprecated" in value.result:
            counts['warnings'] += 1
    
    # Additional analysis of whether failMandatoryExist occurred in the scheme and adding the number of failMandatoryExist to FAILS
    if 'failMandatoryExist' in counts.keys():
        counts['fails'] += counts['failMandatoryExist']

    return True, counts, results, redfish_obj.getLinks(collection_limit), redfish_obj


def validateURITree(service, URI, uriName, expectedType=None, expectedJson=None, parent=None, all_links_traversed=None, inAnnotation=False):
    # from given URI, validate it, then follow its links like nodes
    #   Other than expecting a valid URI, on success (real URI) expects valid links
    #   valid links come from getAllLinks, includes info such as expected values, etc
    #   as long as it is able to pass that info, should not crash
    # If this is our first called URI
    top_of_tree = all_links_traversed is None
    if top_of_tree:
        all_links_traversed = set()
    all_links_traversed.add(URI)

    # Links that are not direct, usually "Redundancy"
    referenced_links = []

    if inAnnotation and service.config['uricheck']:
        service.catalog.flags['ignore_uri_checks'] = True
    validateSuccess, counts, results, gathered_links, thisobj = validateSingleURI(service, URI, uriName, expectedType, expectedJson, parent)
    if inAnnotation and service.config['uricheck']:
        service.catalog.flags['ignore_uri_checks'] = False

    # If successful and a MessageRegistryFile...
    if validateSuccess and 'MessageRegistryFile.MessageRegistryFile' in thisobj.Type.getTypeTree():
        # thisobj['Location'].Collection[0]['Uri'].Exists
        if 'Location' in thisobj:
            if thisobj['Location'].IsCollection:
                val_list = thisobj['Location'].Value
            else:
                val_list = [thisobj['Location'].Value]
            for sub_obj in val_list:
                if 'Uri' in sub_obj:
                    gathered_links.append(sub_obj)

    # If successful...
    if validateSuccess:
        # Bring Registries to Front if possible

        for link in sorted(gathered_links, key=lambda link: (link.Type.fulltype != 'Registries.Registries')):
            if link is None or link.Value is None:
                my_logger.warning('Link is None, does it exist?')
                continue

            # get Uri or @odata.id
            if not isinstance(link.Value, dict):
                errmsg = '{} is expected to be an object containing @odata.id'.format(link.Name)
                my_logger.error(errmsg)
                results[uriName]['errors'] += '\n' + errmsg
                counts['errorInvalidReferenceObject'] += 1
                continue
            link_destination = link.Value.get('@odata.id', link.Value.get('Uri'))

            if link.IsExcerpt or link.Type.Excerpt:
                continue
            if not service.config['oemcheck']:
                if link_destination and '/Oem/' in link_destination or link and 'Resource.OemObject' in link.Type.getTypeTree():
                    my_logger.info('Oem link skipped: {}'.format(link_destination))
                    counts['skipOemLink'] += 1
                    continue
            if any(x in str(link.parent.Type) or x in link.Name for x in ['RelatedItem', 'Redundancy', 'Links', 'OriginOfCondition']) and not link.IsAutoExpanded:
                referenced_links.append((link, thisobj))
                continue
            if link_destination in all_links_traversed:
                counts['repeat'] += 1
                continue
            elif link_destination is None:
                errmsg = 'URI for NavigationProperty is missing {}'.format(uriName)
                my_logger.error(errmsg)
                results[uriName]['errors'] += '\n' + errmsg
                counts['errorMissingOdata'] += 1
                continue
            elif link_destination.split('#')[0].endswith('/'):
                # (elegantly) add warn message to resource html
                warnmsg = 'URI acquired ends in slash: {}'.format(link_destination)
                my_logger.warning(warnmsg)
                results[uriName]['warns'] += '\n' + warnmsg
                counts['warnTrailingSlashLink'] += 1
                newLink = ''.join(link_destination.split('/')[:-1])
                if newLink in all_links_traversed:
                    counts['repeat'] += 1
                    continue

            if link.Type is not None and link.IsAutoExpanded:
                returnVal = validateURITree(service, link_destination, uriName + ' -> ' + link.Name, link.Type, link.Value, thisobj, all_links_traversed, link.InAnnotation)
            else:
                returnVal = validateURITree(service, link_destination, uriName + ' -> ' + link.Name, parent=parent, all_links_traversed=all_links_traversed, inAnnotation=link.InAnnotation)
            success, linkCounts, linkResults, xlinks, xobj = returnVal

            my_logger.verbose1('%s, %s', link.Name, linkCounts)

            referenced_links.extend(xlinks)
            if not success:
                counts['unvalidated'] += 1
            results.update(linkResults)

    if top_of_tree:
        # TODO: consolidate above code block with this
        for link in referenced_links:
            link, refparent = link
            # get Uri or @odata.id
            if link is None or link.Value is None:
                my_logger.warning('Link is None, does it exist?')
                continue
            link_destination = link.Value.get('@odata.id', link.Value.get('Uri'))
            if link.IsExcerpt or link.Type.Excerpt:
                continue
            elif link_destination is None:
                errmsg = 'Referenced URI for NavigationProperty is missing {} {} {}'.format(link_destination, link.Name, link.parent)
                my_logger.error(errmsg)
                results[uriName]['errors'] += '\n' + errmsg
                counts['errorMissingRefOdata'] += 1
                continue
            elif not isinstance(link_destination, str):
                errmsg = 'URI for NavigationProperty is not a string {} {} {}'.format(link_destination, link.Name, link.parent)
                my_logger.error(errmsg)
                results[uriName]['errors'] += '\n' + errmsg
                counts['errorInvalidReferenceObject'] += 1
                continue
            elif link_destination.split('#')[0].endswith('/'):
                # (elegantly) add warn message to resource html
                warnmsg = 'Referenced URI acquired ends in slash: {}'.format(link_destination)
                my_logger.warning(warnmsg)
                results[uriName]['warns'] += '\n' + warnmsg
                counts['warnTrailingSlashRefLink'] += 1
                newLink = ''.join(link_destination.split('/')[:-1])
                if newLink in all_links_traversed:
                    counts['repeat'] += 1
                    continue

            if link_destination not in all_links_traversed:
                my_logger.verbose1('{}, {}'.format(link.Name, link))
                counts['reflink'] += 1
            else:
                continue

            my_link_type = link.Type.fulltype
            success, my_data, _, _ = service.callResourceURI(link_destination)
            # Using None instead of refparent simply because the parent is not where the link comes from
            returnVal = validateURITree(service, link_destination, uriName + ' -> ' + link.Name, my_link_type, my_data, None, all_links_traversed)
            success, linkCounts, linkResults, xlinks, xobj = returnVal
            # refLinks.update(xlinks)

            if not success:
                counts['unvalidatedRef'] += 1
                if 'OriginOfCondition' in link.Name or 'OriginOfCondition' in link.parent.Name:
                    my_logger.info('Link was unsuccessful, but non mandatory')
                else:
                    results.update(linkResults)
            else:
                results.update(linkResults)

    return validateSuccess, counts, results, referenced_links, thisobj



# Copyright Notice:
# Copyright 2016-2024 DMTF. All rights reserved.
# License: BSD 3-Clause License. For full text see link:
# https://github.com/DMTF/Redfish-Service-Validator/blob/main/LICENSE.md

from setuptools import setup
from codecs import open

with open("README.md", "r", "utf-8") as f:
    long_description = f.read()

setup(
    name="redfish_service_validator",
    version="2.4.8",
    description="Redfish Service Validator",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="DMTF, https://www.dmtf.org/standards/feedback",
    license="BSD 3-clause \"New\" or \"Revised License\"",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Topic :: Communications"
    ],
    keywords="Redfish",
    url="https://github.com/DMTF/Redfish-Protocol-Validator",
    packages=["redfish_service_validator"],
    entry_points={
        'console_scripts': [
            'rf_service_validator=redfish_service_validator.RedfishServiceValidator:main',
            'rf_service_validator_gui=redfish_service_validator.RedfishServiceValidatorGui:main'
        ]
    },
    install_requires=[
      "redfish>=3.1.5",
      "requests",
      "beautifulsoup4>=4.6.0",
      "lxml"
    ]
)
